/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <error.h>
#include <stdbool.h>
#include <device-node.h>
#include <vconf.h>
#include <libgdbus/dbus-system.h>

#include "core/log.h"
#include "core/common.h"
#include "core/devices.h"
#include "usb.h"

/* Legacy signals */
#define SIGNAL_STATE_CHANGED  "StateChanged"
#define SIGNAL_MODE_CHANGED   "ModeChanged"
#define SIGNAL_CONFIG_ENABLED "ConfigEnabled"
#define CHANGE_USB_MODE       "ChangeUsbMode"

static int get_usb_current_mode(void)
{
	int val, ret;

	ret = vconf_get_int(VCONFKEY_USB_CUR_MODE, &val);
	if (ret != 0) {
		_E("Failed to get current mode (%d)", ret);
		return ret;
	}

	return val;
}

static int get_usb_state(void)
{
	int ret, val;
	int result = 0;

	ret = vconf_get_int(VCONFKEY_SYSMAN_USB_STATUS, &val);
	if (ret != 0) {
		_E("Failed to get usb_state (%d)", ret);
		return ret;
	}

	if (val == VCONFKEY_SYSMAN_USB_DISCONNECTED)
		result = val;
	else {
		result = VCONFKEY_SYSMAN_USB_CONNECTED;
		if (val == VCONFKEY_SYSMAN_USB_AVAILABLE)
			result |= VCONFKEY_SYSMAN_USB_AVAILABLE;
	}

	return result;
}

/* dbus signals */
void broadcast_usb_config_enabled(int state)
{
	int ret;

	_I("USB config enabled (%d)", state);

	ret = dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_USB,
			DEVICED_INTERFACE_USB, SIGNAL_CONFIG_ENABLED,
			g_variant_new("(i)", state));
	if (ret < 0)
		_E("Failed to send dbus signal");
}

void broadcast_usb_state_changed(void)
{
	int ret;
	unsigned int state;
	static unsigned int prev_state = UINT_MAX;

	state = get_usb_state();
	if (state == prev_state)
		return;
	prev_state = state;

	_I("USB state changed (%u)", state);

	ret = dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_USB,
			DEVICED_INTERFACE_USB, SIGNAL_STATE_CHANGED,
			g_variant_new("(u)", state));
	if (ret < 0)
		_E("Failed to send dbus signal");
}

void broadcast_usb_mode_changed(void)
{
	int ret;
	unsigned int mode;
	static unsigned int prev_mode = UINT_MAX;

	mode = get_usb_current_mode();
	if (mode == prev_mode)
		return;
	prev_mode = mode;

	_I("USB mode changed (%u)", mode);

	ret = dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_USB,
			DEVICED_INTERFACE_USB, SIGNAL_MODE_CHANGED,
			g_variant_new("(u)", mode));
	if (ret < 0)
		_E("Failed to send dbus signal");
}

static void change_usb_client_mode(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)

{
	int req, debug;
	int ret;
	unsigned mode;

	g_variant_get(param, "(i)", &req);

	debug = 0;
	switch (req) {
	case SET_USB_DEFAULT:
	case SET_USB_SDB:
	case SET_USB_SDB_DIAG:
		mode = USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB;
		debug = 1;
		break;
	case SET_USB_RNDIS:
	case SET_USB_RNDIS_DIAG:
	case SET_USB_RNDIS_SDB:
		mode = USB_FUNCTION_RNDIS | USB_FUNCTION_SDB;
		debug = 1;
		break;
	case 11: /* SET_USB_DIAG_RMNET */
	case 12: /* SET_USB_ACM_SDB_DM */
	default:
		_E("(%d) is unknown usb mode", req);
		goto out;
	}

	if (vconf_set_bool(VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL, debug) != 0)
		_E("Failed to set usb debug toggle (%d)", debug);

	ret = usb_change_mode(mode);
	if (ret < 0)
		_E("Failed to change usb mode (%d)", ret);

out:
	return ;
}

/* dbus methods */
GVariant *get_usb_client_state(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int state;

	state = get_usb_state();

	return g_variant_new("(u)", state);
}

GVariant *get_usb_client_mode(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int mode;

	mode = get_usb_current_mode();

	return g_variant_new("(u)", mode);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetState", NULL, "u", get_usb_client_state },  /* from Tizen 2.3 */
	{ "GetMode", NULL, "u", get_usb_client_mode  },   /* from Tizen 2.3 */
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_USB,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};


int usb_dbus_init(void)
{
	int ret;

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_USB, &dbus_interface);
	if (ret < 0)
		_E("fail to init dbus method(%d)", ret);

	ret = subscribe_dbus_signal(NULL, DEVICED_PATH_USB,
		DEVICED_INTERFACE_USB, "ChangeUsbMode",
		change_usb_client_mode, NULL, NULL);
	if (ret <= 0) {
		_E("fail to subscribe dbus signal(%d)", ret);
		return ret;
	}

	return 0;
}
